import re
from operator import itemgetter

import jinja2


def get_values(asset, key, pattern="(*.)"):
    '''
    Returns values extracted from a metadata property of an asset named by key.
    '''
    if isinstance(pattern, basestring):
        pattern = re.compile(pattern)
    source = asset.get(key, '')
    return [v.strip() for v in pattern.findall(source) if v]


def render(path, template, context):
    '''
    Uses Jinja2 to render a template relative to the path with the given context.
    '''
    return jinja2.Environment(
        loader=jinja2.FileSystemLoader(path)
    ).get_template(template).render(context)


def pathurl(path):
    '''
    Returns a nicer URL for the path on disk for webservers that treat URLs missing a
    filename to be requesting index.html

    If the path ends in `/index.html`, make it end with `/`.

    Make every path absolute.
    '''
    if path.endswith("/index.html"):
        path = path.replace("/index.html", "/")
    if not path.startswith("/"):
        path = "/" + path
    return path


def _calculate_totals(aggregates):
    '''
    From a list of Aggregates return a sorted list of tuples containing each Aggregate
    and how many content assets it has.
    '''
    totals = []
    for aggregate in aggregates:
        assets = aggregate['assets']
        totals.append([aggregate, len(assets)])
    return sorted(totals, key=itemgetter(1), reverse=True)


def generate_cloud(items, minsize=0.8, maxsize=1.0):
    '''
    From a list of Aggregates return a list of tuples containing each Aggregate and its
    relative "size" in a tag cloud normalized to 1.0.

    This is useful for rendering into a <span> tag's `style` attribute to affect
    font-size.
    '''
    totals = _calculate_totals(items)
    raw_totals = [i[1] for i in totals]
    max_total = float(max(raw_totals) if raw_totals else 1)
    cloud = []  # generate sizes
    for item, total in totals:
        size = round(((total - 1) / (max_total - 1 or 1)) * (maxsize - minsize) + minsize, 2)
        cloud.append((item, size))
    return cloud



import bs4

from summarize import summarize_blocks


class AutoSummary(object):
    '''
    Generates a summary from an asset's content.
    '''
    def __init__(self, key="summary"):
        self.key = key  # where to store the summary

    def process(self, context):
        for asset in context['assets']:
            # don't generate a summary if one has been manually written
            if asset.get(self.key):
                continue

            html = bs4.BeautifulSoup(asset.content, 'html.parser')
            summaries = summarize_blocks(map(lambda p: p.text, html.find_all('p')[2:3]))
            asset[self.key] = summaries[0]  # get the first summary



import os

from blot.assets import ContentAsset


class StaticReader(object):
    '''
    Simple reader which performs no content or metadata parsing.
    '''
    def __init__(self, asset_class=ContentAsset):
        self.asset_class = asset_class

    def read(self, paths):
        assets = []
        for path in paths:
            with open(path, 'r') as fobj:
                content = fobj.read()
                asset = self.asset_class(path, content, {})
                assets.append(asset)
        return assets



from slugify import slugify

from blot.assets import Aggregator


class Tags(Aggregator):
    '''
    Simple multi-value Aggregator which splits the named metadata property by comma.
    '''
    def __init__(self, key='tags', pattern="([^,]+)"):
        super(Tags, self).__init__(key, pattern)

    def process_aggregate(self, tag, asset):
        tag['name'] = slugify(tag['name'])
        asset_tags = asset.get('tags')
        if not isinstance(asset_tags, list):
            asset['tags'] = []
        asset['tags'].append(tag)

    def finish(self, context, tags):
        context['tags'] = tags.values()



import math

from blot import utils


class PageHelper(object):
    '''
    Utility class that providers an iterator that returns slices of the underlying list.
    '''
    def __init__(self, items, size):
        self.items = items
        self.size = size
        self.length = math.ceil(len(items) / float(size))

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        index = float(index)
        if index >= self.length:
            raise IndexError("PageHelper has no {} page.".format(index))
        start = int(index * self.size)
        return self.items[start:start + self.size]

    def __iter__(self):
        index = 0
        while index < self.length:
            yield self[index]
            index += 1


class PaginatedWriter(object):
    '''
    Writer that paginates the underlying assets and renders a template for each page.

    PaginatedWriter takes divides its assets into groups of the specified size and
    renders the specified template for each group. The global context variable will
    point to the URL of the first page.

    The context when rendering each of the pages contains a number of variables for
    navigating between the pages.
    '''

    def __init__(self, assets, variable_name, template, path, size=10):
        self.assets = assets
        self.variable_name = variable_name
        self.template = template
        self.path = path
        self.size = size
        self.helper = PageHelper(assets, size)

    def pathfor(self, index):
        '''generate the target path for a given page'''
        postfix = index + 1 if index > 0 else ''
        return self.path.format(page=postfix)

    def target(self, context):
        path = self.pathfor(0)  # get path for first page
        context[self.variable_name] = utils.pathurl(path)

    def render(self, original_context):
        template_path = original_context.get('TEMPLATE_PATH', './')
        helper = PageHelper(self.assets, self.size)
        for index, assets in enumerate(helper):
            context = original_context.copy()
            context['assets'] = assets
            context['page_number'] = index + 1
            context['first_page'] = self.pathfor(0)
            context['last_page'] = self.pathfor(len(self.helper) - 1)
            if index > 0:
                context['previous_page'] = self.pathfor(index - 1)
            else:
                context['previous_page'] = None
            if index < len(self.helper) - 1:
                context['next_page'] = self.pathfor(index + 1)
            else:
                context['next_page'] = None
            path = self.pathfor(index)
            output = utils.render(template_path, self.template, context)
            yield (path, output)





from .autosummary import AutoSummary
from .base import ContentAsset, Aggregator, Aggregate
from .categories import Categories
from .categorytags import CategoryTags
from .humanizer import Humanizer
from .metadata import PathMetadata
from .series import Series
from .seriestags import SeriesTags
from .slugifier import Slugifier
from .tags import Tags
from .titler import Titler



import re
from operator import itemgetter


from titlecase import titlecase

from blot.utils import get_values
from blot.assets import Aggregator


class Series(Aggregator):
    '''
    Aggregator which attempts to automatically discover multi-part content series.

    Like other Aggregators the Series processor generates a new set of assets based on
    values extracted from a specified key of some target assets. However, the Series
    processor will only generate new assets for aggregate values found across two or
    more target assets.

    For each Series generated a number of metadata properties are added to target
    assets. These properties contain the first, last, previous and next target assets
    in the Series. This allows a theme to construct navigation between each asset in
    the Series.

    In addition to the normal aggregation metadata property, the Series processor
    requires direction for deriving the order of assets in the Series. So the
    `part_key` and `part_pattern` are used to select this value from the metadata of
    target assets. If the extracted part value is 'index.html' it will automatically be
    converted to 1, putting it as the first part in the series.

    '''

    def __init__(self, key='parent', pattern="(.*)",
                 part_key='filename', part_pattern="(.*)"):
        super(Series, self).__init__(key, pattern)
        self.part_key = part_key
        self.part_pattern = re.compile(part_pattern)

    def process_aggregate(self, series, asset):
        '''generate a title for the series itself'''
        title = series['name'].replace("-", " ")
        title = title.replace("_", " ")
        series['title'] = titlecase(title)

    def get_part_number(self, asset):
        '''derive the part number of an asset'''
        part_number = get_values(asset, self.part_key, self.part_pattern)[0]
        if part_number == 'index':
            return 1
        else:
            return int(part_number)

    def finish(self, context, series):
        # only aggregates containing 2 or more assets are valid series
        valid_series = []
        for name, series_obj in series.items():
            assets = series_obj['assets']
            asset_count = len(assets)
            if asset_count > 1:
                # if the series is actually valid, then update its assets
                for asset in assets:
                    asset['series'] = series
                    asset['part_count'] = asset_count
                    asset['part_number'] = self.get_part_number(asset)
                # once part numbers are known, sort the series assets
                assets.sort(key=itemgetter('part_number'))
                # add relative parts for navigation
                for index, asset in enumerate(assets):
                    asset['first_part'] = assets[0]
                    asset['last_part'] = assets[-1]
                    asset['next_part'] = assets[(index + 1) % asset_count]
                    asset['previous_part'] = assets[index - 1]
                valid_series.append(series_obj)
        # add the series to the target content type's context
        context['series'] = valid_series



import dateparser
import humanize


class Humanizer(object):
    '''
    Processor which runs `naturaltime` over the specified metadata property.
    '''
    def __init__(self, attr='date'):
        super(Humanizer, self).__init__()
        self.attr = attr

    def process(self, context):
        for asset in context['assets']:
            value = asset[self.attr]
            datetime = dateparser.parse(value)
            asset[self.attr] = humanize.naturaltime(datetime)



class StaticWriter(object):
    '''
    Writer that writes asset content without rendering.
    '''

    def __init__(self, assets, path):
        self.assets = assets
        self.path = path

    def target(self, context):
        for asset in self.assets:
            asset.target = self.path.format(**asset.metadata)

    def render(self, context):
        context['assets'] = self.assets
        for asset in self.assets:
            context['asset'] = asset
            yield (asset.target, asset.content)



import os


class PathMetadata(object):
    '''
    Generates a number of metadata properties based on an asset's source path.
    '''
    def process(self, context):
        for asset in context['assets']:
            md = asset.metadata
            dirname, basename = os.path.split(asset.source)
            filename, extension = os.path.splitext(basename)
            ancestry, parent = os.path.split(dirname)
            md.update(dict(
                dirname=dirname, basename=basename,
                filename=filename, extension=extension,
                ancestry=ancestry, parent=parent,
            ))



class CategoryTags(object):
    '''
    Processor that adds to each Category for a content type, all the Tags of its assets.
    '''
    def process(self, context):
        for asset in context['assets']:
            category = asset.get('category')
            tags = asset.get('tags')
            if category and tags:
                category['tags'] = category.get('tags', [])
                for tag in tags:
                    if tag not in category['tags']:
                        category['tags'].append(tag)



import os


def _generate_type_context(type_config):
    '''
    Generate a context dictionary specific to a Content-Type.

    Takes a content type configuration dictionary and produces
    a content type specific context for rendering. This
    involves loading and reading any discoverable content
    sources and applying all of the configured asset processors
    against each.

    The resulting context's `assets` key will contain a list
    of the processed Content Asset objects. Other keys may be
    populated by Processors with secondary Processed Assets.

    Processors may also add keys not containing assets.
    '''
    context = {}
    loader = type_config['loader']
    reader = type_config['reader']
    processors = type_config['processors']

    # discover content with loader
    content_paths = loader.load()
    # parse source for content and metadata
    context['assets'] = reader.read(content_paths)
    # process content assets and update context
    for processor in processors:
        processor.process(context)
    return context


def read(context, content_types):
    '''
    Stage 1 of site generation.

    For each defined content type:
      - use loader to discover input content
      - use reader to parse input content into content assets
      - run each processor against the assets
      - return resulting context dictionary containing everything

    '''
    for type_name, type_config in content_types.items():
        # each content type produces a distinct context which is
        # stored under the name of the content type in the global
        # rendering context
        context[type_name] = _generate_type_context(type_config)
    # return the now asset populated global context
    return context


def write(context, writers, build_path="output"):
    '''
    Stage 2 of site generation.

    For each specified writer:
      - Resolve each asset's target output location on disk

    For each specified writer:
      For each asset fed to the writer:
        - Render the asset using the global context
        - Write the asset to disk

    The reason for targetting all assets before rendering is so that during rendering
    the URLS for all assets are available. This allows any templates being used to
    render a specific content type to include links to other assets of other content
    types.

    '''
    # have each Writer target all of its Content Assets
    for writer in writers:
        writer.target(context)

    # have each Writer render and write all of its Content Assets
    for writer in writers:
        context_copy = context.copy()  # rendering should be safe
        for target, content in writer.render(context_copy):
            full_target = os.path.join(build_path, target)
            path, _ = os.path.split(full_target)

            if not os.path.exists(path):
                os.makedirs(path)

            with open(full_target, 'w') as fobj:
                fobj.write(content)



import os

import markdown2

from blot.assets import ContentAsset


class MarkdownReader(object):
    '''
    Reader which parses a content source as Markdown.

    Metadata should be specified by a "fenced" yaml block at the top of the
    file. Metadata keys will be lower-cased. Duplicate keys will override ones earlier
    in the file.

    '''
    def __init__(self,
                 asset_class=ContentAsset,
                 extras=['metadata', 'fenced-code-blocks']):
        self.asset_class = asset_class
        self.extras = extras

    def read_path(self, path):
        with open(path, 'r') as fobj:
            data = fobj.read()
            content = markdown2.markdown(data, extras=self.extras)
            metadata = {k.lower(): v for k, v in content.metadata.items()}
            return self.asset_class(path, content, metadata)

    def read(self, paths):
        assets = []
        for path in paths:
            asset = self.read_path(path)
            assets.append(asset)
        return assets



from blot.assets import Aggregator


class Categories(Aggregator):
    '''
    A simple single value Aggregator for organizing content assets.
    '''

    def __init__(self, key='category', pattern='(.*)'):
        super(Categories, self).__init__(key, pattern)

    def process_aggregate(self, category, asset):
        # store the derived aggregate onto the `category` property of the asset
        asset['category'] = category

    def finish(self, context, categories):
        # make the list of categories aggregated for this content type available in
        # the content type's context under `categories`
        context['categories'] = categories.values()



class SeriesTags(object):
    '''
    Processor that adds to each Series for a Content Type, all the Tags of its assets.
    '''
    def process(self, context):
        for asset in context['assets']:
            series = asset.get('series')
            tags = asset.get('tags')
            if series and tags:
                series['tags'] = series.get('tags', [])
                for tag in tags:
                    if tag not in series['tags']:
                        series['tags'].append(tag)



from titlecase import titlecase


class Titler(object):
    '''
    Processor that titlecases an asset metadata property.
    '''
    def __init__(self, source_attr):
        super(Titler, self).__init__()
        self.source_attr = source_attr

    def process(self, context):
        for asset in context['assets']:
            md = asset.metadata
            title = md[self.source_attr]
            title = title.replace("-", " ")
            title = title.replace("_", " ")
            md['title'] = titlecase(title)



from blot import utils


class ListWriter(object):
    '''
    Writer that renders a jinja2 template for each targetted asset.
    '''
    def __init__(self, assets, template, path):
        self.assets = assets  # targetted assets
        self.template = template  # filename of template to use
        self.path = path  # filesystem path pattern to render to

    def target(self, context):
        # render the path pattern with each asset's metadata
        for asset in self.assets:
            # set the result on each asset
            asset.target = self.path.format(**asset.metadata)

    def render(self, context):
        # render each of the assets out to its target
        path = context.get('TEMPLATE_PATH', './')
        context['assets'] = self.assets
        for asset in self.assets:
            context['asset'] = asset
            output = utils.render(path, self.template, context)
            yield (asset.target, output)



from blot import utils


class IndexWriter(object):
    '''
    Writer that renders a single jinja2 template. The global context is updated with a
    URL to the rendered file during targetting. All assets are passed into the context
    as `assets` during rendering.

    '''
    def __init__(self, assets, variable_name, template, path):
        self.assets = assets  # targetted assets
        self.variable_name = variable_name  # global context variable for URL
        self.template = template  # filename of template to use
        self.path = path  # filesystem path to write to

    def target(self, context):
        # update global context with URL to this index
        context[self.variable_name] = utils.pathurl(self.path)

    def render(self, context):
        path = context.get('TEMPLATE_PATH', './')
        # add this index's assets to the context
        context['assets'] = self.assets
        output = utils.render(path, self.template, context)
        yield (self.path, output)




#!/usr/bin/env python

from setuptools import setup

setup(name="blot",
      version="0.0.8",
      packages=['blot', 'blot.readers', 'blot.assets', 'blot.writers'],
      package_dir={
          "blot": "blot",
          "blot.readers": "blot/readers",
          "blot.assets": "blot/assets",
          "blot.writers": "blot/writers"
      },
      install_requires=[
          "nltk==3.2.1",
          "dateparser==0.4.0",
          "awesome-slugify==1.6.5",
          "bs4==0.0.1",
          "Fabric==1.11.1",
          "humanize==0.5.1",
          "Jinja2==2.8",
          "mock==2.0.0",
          "nose==1.3.7",
          "Pygments==2.1.3",
          "titlecase==0.8.1",
          "markdown2==2.3.1",
          "summarize_github==0.0.1"
      ],
      dependency_links=[
          'http://github.com/vgel/summarize.py/tarball/0911f23a500dcb0de7b2722587f2dff3598a443a#egg=summarize_github-0.0.1',
      ],
      # metadata for upload to PyPI
      author="Dustin Lacewell",
      author_email="dlacewell@gmail.com",
      description="An unassuming static site-generator for Python",
      keywords="site-generator",
      url="http://github.com/dustinlacewell/blot", )



from slugify import Slugify


class Slugifier(object):
    '''
    Processor that slugifies an asset metadata property.
    '''
    def __init__(self, source_attr='title'):
        super(Slugifier, self).__init__()
        self.slugifier = Slugify(to_lower=True)
        self.source_attr = source_attr

    def process(self, context):
        for asset in context['assets']:
            md = asset.metadata
            md['slug'] = self.slugifier(md[self.source_attr])



from re import compile
from operator import itemgetter

from blot.utils import pathurl, get_values


class ContentAsset(object):
    '''An object with some data and associated metadata.

    ContentAssets represent the result of loading and reading content sources. The
    result of parsing the source source provides its content and optional metadata
    dictionary.

    ContentAssets are fed through a pipeline of processors where they may be modified
    in various ways. Eventually, they are passed to Writers which can render them to
    their final destinations on disk.

    A basic dictionary-like access is avaiable as a shortcut to the asset's
    metadata. Once the assets's `target` attribute has been set, its `url` property
    will become available.

    Some assets are produced by the processing of other assets and have no source.
    '''
    def __init__(self, source=None, content=None, metadata={}):
        self.content = content  # data of the asset
        self.source = source  # where the asset came from (usually a file path)
        self.metadata = metadata  # initial metadata
        self.target = None  # where this asset will live on disk during write

    def __getitem__(self, key):
        return self.metadata[key]

    def __setitem__(self, key, value):
        self.metadata[key] = value

    def get(self, key, default=None):
        return self.metadata.get(key, default)

    @property
    def url(self):
        # a url to the asset based on where it lives on disk
        return pathurl(self.target)


class Aggregate(ContentAsset):
    '''
    A special type of ContentAsset that has two known metadata properties:
      - name : The metadata value of some other ContentAsset this Aggregate is based on
      - assets : The list of ContentAssets in which this Aggregate appears
    '''

    def __init__(self, name, assets=None):
        super(Aggregate, self).__init__(
            None, None, dict(name=name, assets=assets or []))

    def __str__(self):
        return self.get('name', '')

    def __repr__(self):
        return str(self)


class Aggregator(object):
    '''Base-class for an asset processor that generates Aggregates for each value in the
    metadata of processed ContentAssets.

    When used as an asset processor, it will look at a specific metdata property of
    each ContentAsset. For each unique value found at this property, a new Aggregate
    asset will be generated.

    Each Aggregate object contains in its own metadata an `assets` key that contains a
    list of each of the ContentAssets in which the aggregated value was
    found.

    It is up to subclasses of Aggregator to make the list of generated Aggregate assets
    available. Subclasses also have the chance to update each aggregate and the asset
    it comes from.

    Generally, the original content type's context will be updated with a list of the
    Aggregates under a relevant key in the `finish` method. Check the Tags and
    Categories processors for examples.
    '''

    def __init__(self, key, pattern, asset_class=Aggregate):
        self.key = key  # metadata key to aggregate over
        self.pattern = compile(pattern)  # regular expression to extract aggregate values
        self.asset_class = asset_class  # class from which to create new aggregate assets

    def get_values(self, asset):
        '''
        Get the aggregated property value from the asset.
        '''
        return get_values(asset, self.key, self.pattern)

    def process_aggregate(self, aggregate, asset):
        pass

    def finish(self, context, aggregates):
        pass

    def process(self, context):
        '''
        Check each asset in the content Type context and aggregate discovered values
        over the specific metadata key into new assets.
        '''
        aggregates = dict()
        for asset in context['assets']:
            values = self.get_values(asset)
            for value in values:
                aggregate = aggregates.get(value) or self.asset_class(value)
                aggregate['assets'].append(asset)
                self.process_aggregate(aggregate, asset)
                aggregates[value] = aggregate
        self.finish(context, aggregates)
